import pyopencl as cl
import os
import numpy as np

def zero_default_opencl_context():
    os.environ["PYOPENCL_CTX"] = "0"

def run_kernel(name, shape, *, inputs = None, outputs_number = None, ctx=[], queue = [], prg = [], source_code = None, file_name = 'compute_shaders.cl'):
    """
    Runs OpenCL kernel
    :param name: str = name of kernel
    :param shape: Tuple[int, int] = shape of all inputs and outputs
    :param inputs: list = list of input frames
    :param outputs_number: int = number of elements in the output
    :param ctx: Cache - don't change
    :param queue: Cache - don't change
    :param prg: Cache - don't change
    :return: list = list of outputs (images)
    """
    if inputs is None or outputs_number is None:
        raise(Exception(f"inputs = \'{inputs}\' or outputs_number = \'{outputs_number}\' is None"))
    if len(ctx)==0:
        ctx.append(cl.create_some_context())
        queue.append(cl.CommandQueue(ctx[0]))
        cl_code_source = ""
        if source_code is None:
            with open(file_name, 'r', encoding="utf8") as file:
                cl_code_source = file.read()
        else:
            cl_code_source = source_code
        prg.append(cl.Program(ctx[0], cl_code_source).build())

    h = shape[0]
    w = shape[1]
    src_bufs = [cl.image_from_array(ctx[0], i, 4) for i in inputs]
    fmt = cl.ImageFormat(cl.channel_order.RGBA, cl.channel_type.UNSIGNED_INT8)
    dest_bufs = [cl.Image(ctx[0], cl.mem_flags.WRITE_ONLY, fmt, shape=(w, h)) for i in range(outputs_number)]
    # execute OpenCL function
    prg[0].__getattr__(name)(queue[0], (w, h), None, *src_bufs, *dest_bufs)
    # copy result back to host
    res=[]
    for i in range(outputs_number):
        dest = np.empty_like(inputs[0])
        cl.enqueue_copy(queue[0], dest, dest_bufs[i], origin=(0, 0), region=(w, h))
        res.append(dest)
    return res


from functools import wraps
import inspect
import copy

def ignore(item):
    return ("WvK43p#AK6n3&8nIh#mLI9rhb$I5nmir3$lp0Kzq*UGvta9#RE_Slava_Ukraini",item)

def reinit_default_args(func):
    """
    decorator that creates new default arguments on each function call.
    Usage:

    to make new default arg on each call simply use:

    @reinit_default_args
    def func(a=[]):
        ...

    to ignore some default values use:
    @reinit_default_args
    def func(a=ignore([])):
        ...
        
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        args_func = inspect.getfullargspec(func)
        print(args_func)
        if args_func.defaults is not None:
            for i in range(len(args_func.defaults)):
                i=-1-i
                if args_func.args[i] not in kwargs:
                    if isinstance(args_func.defaults[i], tuple):
                        if len(args_func.defaults[i])==2:
                            if args_func.defaults[i][0]=="WvK43p#AK6n3&8nIh#mLI9rhb$I5nmir3$lp0Kzq*UGvta9#RE_Slava_Ukraini":
                                kwargs[args_func.args[i]] = args_func.defaults[i][1]
                                continue
                    kwargs[args_func.args[i]] = copy.copy(args_func.defaults[i])
        if args_func.kwonlyargs is not None:
            for i in range(len(args_func.kwonlyargs)):
                if args_func.kwonlyargs[i] not in kwargs:
                    if isinstance(args_func.kwonlydefaults[args_func.kwonlyargs[i]], tuple):
                        if len(args_func.kwonlydefaults[args_func.kwonlyargs[i]])==2:
                            if args_func.kwonlydefaults[args_func.kwonlyargs[i]][0]=="WvK43p#AK6n3&8nIh#mLI9rhb$I5nmir3$lp0Kzq*UGvta9#RE_Slava_Ukraini":
                                kwargs[args_func.kwonlyargs[i]] = args_func.kwonlydefaults[args_func.kwonlyargs[i]][1]
                                continue
                    kwargs[args_func.kwonlyargs[i]] = copy.copy(args_func.kwonlydefaults[args_func.kwonlyargs[i]])
        func(*args, **kwargs)
    return wrapper_func